/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.text.ParseException;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.ConfigFileException;
import org.ucam.srcf.assassins.exceptions.PluginInstantiationException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.exceptions.XMLParseException;
import org.ucam.srcf.assassins.xml.Config;
import org.ucam.srcf.assassins.xml.WebPageParser;


/**
 * The main class that actually takes a command and does things with it.
 */
public class autoumpire
{
   private static final String SYNTAX = "Syntax: autoumpire [options] command [command option]\n"
      +" Options:\n"
      +"\t-f <configfile>\tSpecify a config file - defaults to config.xml\n"
      +"\t-t <time>\tTime to run the statebuilder at - defaults to now\n";
   
   private static boolean natv;
   private static native void realexecnative(String cmd);

   static {
      try {
         System.load("./autoumpire.so");
         natv = true;
      } catch(Throwable t) {
         natv = false;
      }
   }
   private static void execnative(String cmd)
   {
      if (natv) realexecnative(cmd);
      else
         System.out.println("Not doing exec(), native library not loaded");
   }

   /*
    * Setup map of parameters -> commands
    */
   public static void setupCommandMap(Map commands)
   {
      commands.put("add-player", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException { addPlayer(plugins, params); } } );
      
      commands.put("add-assassin", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException { addAssassin(null,plugins, params); } } );
      
      commands.put("edit-assassin", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException { editAssassin(plugins, params); } } );
      
      commands.put("add-event", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  addEvent(plugins, params); } } );
      
      commands.put("add-participant", new Command() { public void command(Map plugins, Map params, long nowtime) 
      throws AssassinSQLException, IOException, ParseException {  addParticipant(plugins, params); } } );
      
      commands.put("add-event-with-report", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException 
            {  
               Event e = addEvent(plugins,params);
               if (null==e) {
                  return;
               }
               addReport(e,plugins,params);
            } } );
      
      commands.put("edit-event", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException { editEvent(plugins, params); } } );
      
      commands.put("add-report", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  addReport(null,plugins, params); } } );
      
      commands.put("edit-report", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  editReport(plugins, params); } } );
      
      commands.put("add-pseudonym", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  addPseudonym(plugins, params); } } );      
      
      commands.put("add-player-with-assassin", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException 
            {  
               Player p = addPlayer(plugins,params);
               addAssassin(p,plugins,params);
            } } );
      
      commands.put("dump-state", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  dumpState(plugins, params); } } );      

      commands.put("update-state", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  updateState(plugins, params); } } );      

      commands.put("generate-mail", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException {  generateMail(plugins, params); } } );      
      
      commands.put("generate-pages", new Command() { public void command(Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException, StateBuilderException, XMLParseException {  generatePages(plugins, params); } } );

      commands.put("dump-killtree", new Command() { public void command(Map plugins, Map params, long nowtime)
	      throws AssassinSQLException, IOException, ParseException, StateBuilderException, XMLParseException { dumpKillTree(plugins, params); } } );
   }
   
   public static void main(String[] args)
   {
      Map commands = new HashMap();

      try {
         /*
          * Read Args
          */
         String cfgfilename = "config.xml";
         String cmd = null;
         Map globalParams = new HashMap();
         Map commandParams = new HashMap();
         Map params = globalParams;
         int argpos = 0;
         while (argpos < args.length) {
            if (Debug.debug) Debug.print(autoumpire.class, "Found argument: "+args[argpos]);
            // its an option               
            if (args[argpos].startsWith("-")) {
               if (argpos == args.length-1) params.put(args[argpos], "");
               else if (args[argpos+1].startsWith("-")) 
                  params.put(args[argpos], "");
               else
                  params.put(args[argpos], args[++argpos]);
               if (Debug.debug) Debug.print(autoumpire.class, "Argument Map: "+params);
            }

            // its not, must be the command
            else {
               if (null == cmd) cmd = args[argpos];
               params = commandParams;
            }
            
            argpos++;               
         }

         /*
          * Check special options
          */
         if (globalParams.containsKey("-v") || globalParams.containsKey("--version"))
         {
            System.out.println(Tools.VERSIONSTRING);
            System.exit(0);
         }
         if (globalParams.containsKey("-h") || globalParams.containsKey("--help"))
         {
            System.err.println(SYNTAX);
            System.out.println("Run with no parameters for a list of commands");
            System.exit(0);
         }
      
         /*
          * Init config file
          */
         try
         {
            if (null != globalParams.get("-f"))
               Config.init((String) globalParams.get("-f"));
            else
               Config.init(cfgfilename);
         }
         catch (ConfigFileException CFe)
         {
            System.err.println("autoumpire: Error reading config file.");
            System.err.println(CFe.getMessage());
            System.err.println(SYNTAX);
            System.exit(1);
         }

         /*
          * Init database connection
          */
         try
         {    
            DatabaseBackend.init();
         }
         catch (AssassinSQLException ASe)
         {
            System.err.println("autoumpire: Error initialising database.");
            System.err.println(ASe.getMessage());
            System.exit(1);
         }

         /*
          * Instantiate and init plugins
          */
         Map pluginnames = Config.getPluginList();
         if (Debug.debug) Debug.print(autoumpire.class, "Got "+pluginnames.size()+" plugins");
         Iterator i = pluginnames.keySet().iterator();
         Map plugins = new HashMap();
         try
         {
            try 
            {
               while (i.hasNext())
               {
                  String pname = (String) i.next();
                  String classname = (String) pluginnames.get(pname);
                  if (null != classname && !("".equals(classname)))
                  {
                     if (Debug.debug) Debug.print(autoumpire.class, "instantiating " + pname);
                     Class c = Class.forName(classname);
                     Plugin p = (Plugin) c.newInstance();
                     if (Debug.debug) Debug.print(autoumpire.class, "setting name");
                     p.setPluginName(pname);
                     if (Debug.debug) Debug.print(autoumpire.class, "setting connector");
                     p.setPluginConnector(new PluginConnector());
                     if (Debug.debug) Debug.print(autoumpire.class, "getting new commands");
                     Map m = p.getCommands();                     
                     if (Debug.debug) Debug.print(autoumpire.class, "Got "+m);
                     commands.putAll(m);
                     plugins.put(pname, p);
                  }
               }
            }
            catch (InstantiationException Ie)
            { if (Debug.debug) Debug.print(autoumpire.class, "Exception has occurred: "+Ie); throw new PluginInstantiationException(Ie); }
            catch (IllegalAccessException IAe)
            { if (Debug.debug) Debug.print(autoumpire.class, "Exception has occurred: "+IAe); throw new PluginInstantiationException(IAe); }
            catch (ClassNotFoundException CNFe)
            { if (Debug.debug) Debug.print(autoumpire.class, "Exception has occurred: "+CNFe); throw new PluginInstantiationException(CNFe); }
            catch (ClassCastException CCe)
            { if (Debug.debug) Debug.print(autoumpire.class, "Exception has occurred: "+CCe); throw new PluginInstantiationException(CCe); }
         }
         catch (PluginInstantiationException PIe)
         {
            System.err.println("autoumpire: Could not instantiate plugin!");
            System.err.println(PIe.toString());
            System.err.println(PIe.getMessage());
            System.exit(1);
         }

         /*
          * Add standard commands to map
          */

         setupCommandMap(commands);

         /*
          * Get Time to calculate on
          */
         String tstr = (String) globalParams.get("-t");
         if (null == tstr)
            Tools.setNowTime(new Date());
         else
            Tools.setNowTime(tstr);

          if (Debug.debug) Debug.print(autoumpire.class, "Acting as time "+Tools.formatTime(Tools.getNowTime()));

         /*
          * Do command
          */
         
         Command com = (Command) commands.get(cmd);
         if (null != com) {
            if (Debug.debug) Debug.print(autoumpire.class, "Running Command "+cmd);
            try {
            	com.command(plugins, commandParams, Tools.getNowTime());
            	DatabaseBackend.commitAll();
            } 
            catch (ParseException Pe)
            {
               System.err.println("autoumpire: Error parsing data while doing "+cmd);
               System.err.println(Pe.getMessage());
               System.exit(1);
            }
            catch (AssassinSQLException ASe)
            {
               System.err.println("autoumpire: SQL error while doing "+cmd);
               System.err.println(ASe.getMessage());
               System.exit(1);
            }
            catch (IOException IOe)
            {
               System.err.println("autoumpire: I/O error while doing "+cmd);
               System.err.println(IOe.getMessage());
               System.exit(1);
            }
         }

         // invalid command
         else {
            System.err.println(SYNTAX);
            System.err.println("Commands:");
            Iterator j = commands.keySet().iterator();
            while (j.hasNext())
               System.err.println("\t"+(String) j.next());
            System.exit(1);
         }

      }
      catch(Exception x)
      {
         System.err.println("autoumpire: exception");
         x.printStackTrace();
         System.err.println(x.toString());
         System.err.println(x.getMessage());
         System.exit(1);
      }
   }

   static Map viewerifyStateMap(Map in)
   {
      Map out = new HashMap();
      Iterator i = in.keySet().iterator();
      while (i.hasNext())
      {
         Assassin key = (Assassin) i.next();
         AssassinState value = (AssassinState) in.get(key);
         if (Debug.debug)
         {
            if (null == key || null == value)
               Debug.print(autoumpire.class, "null key or value in viewerifyStateMap");
         }
         AssassinViewer av = new AssassinViewer(key);
         AssassinStateViewer asv = new AssassinStateViewer(value);
      //         if (Debug.debug) Debug.print(autoumpire.class, "inserting into new map: " + av + "," + asv);
         out.put(av, asv);
      }
      return out;
   }

   static String sendEmails(Map state, Map plugins, boolean force, String manualText) throws AssassinSQLException, ParseException
   {
      if (Debug.debug) Debug.print(autoumpire.class,"sendEmails("+state+","+plugins+","+force+","+manualText+")");
      String emails = "";
      String from = Config.getConfig("vanilla", "from-address");
      String subject = Config.getConfig("vanilla", "email-subject");
      Iterator i = state.values().iterator();
      Map statev = viewerifyStateMap(state);
      while(i.hasNext())
      {
         AssassinState as = (AssassinState) i.next();
         if (Debug.debug) Debug.print(autoumpire.class,""+as);
         if (!as.isAlive()) continue;
         if (!as.isActive()) continue;
         boolean needsemail = force;
         Iterator j = plugins.values().iterator();
         AssassinStateViewer asv = new AssassinStateViewer(as);
         while(j.hasNext() && !needsemail) 
            needsemail |= ((Plugin) j.next()).playerNeedsEmail(asv);


         if (needsemail) {
            Player p = as.getAssassin().getPlayer();
            // get text
            String emailText = "Your Details:\n";
            emailText += "           Name: "+p.getName()+"\n";
            emailText += "        Address: "+p.getAddress()+"\n";
            emailText += "        College: "+p.getCollege()+"\n";
            emailText += "   Water Status: "+p.getWaterStatus()+"\n";
            emailText += "          Notes: "+p.getNotes()+"\n";

            j = plugins.values().iterator();
            while(j.hasNext()) {
               String emailadd = ((Plugin) j.next()).generateEmail(asv, statev);
               if (null != emailadd) emailText += "\n----------------\n" + emailadd;
            }

            // print text & headers
            emails += "MAIL FROM:" + from + "\n";
            emails += "RCPT TO:" + p.getEmail() + "\n";
            emails += "DATA\n";
            emails += "From: "+ from +"\n";
            emails += "Subject: "+ subject +"\n";
            emails += manualText + "\n";
            emails += emailText + "\n\n.\n";
         }
      }
      return emails;
   }

   static void dumpState(Map plugins, Map params) throws AssassinSQLException
   {
      PrintStream out = System.out;
   	StateBuilder sb = new StateBuilder(plugins);
      Map state;
      try {
         state = sb.buildStateAtTime(Tools.getNowTime());
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      Iterator i = state.values().iterator();
      while (i.hasNext())
      {
         AssassinState s = (AssassinState) i.next();
         Assassin a = s.getAssassin();
         Player p = a.getPlayer();
         out.print("*** Assassin played by ");
         out.println(p.getName());
         Iterator pni = s.getKnownPlugins().iterator();
         while (pni.hasNext())
         {
            String pname = (String) pni.next();
            Map pstate = s.getState(pname);
            Iterator ki = pstate.keySet().iterator();
            while (ki.hasNext())
            {
               String field = (String) ki.next();
               out.print(pname);
               out.print(", ");
               out.print(field);
               out.print(" => ");
               Object v = s.getProperty(pname, field);
               out.println(null == v ? "null" : v.toString());
            }
         }
         Iterator psi = s.getPseudonyms().iterator();
         while (psi.hasNext())
         {
            Pseudonym ps = (Pseudonym) psi.next();
            System.out.println(ps.getValue());
         }
      }
   }

   static Player addPlayer(Map plugins, Map params) throws AssassinSQLException, IOException
   {
      Vector qs = new Vector();
      TextQuestion name = new TextQuestion("Real Name:", "");
      qs.add(name);
      TextQuestion email = new TextQuestion("Email:", "");
      qs.add(email);
      TextQuestion college = new TextQuestion("College:","");
      qs.add(college);
      TextQuestion address = new TextQuestion("Address:","");
      qs.add(address);
      TextQuestion water = new TextQuestion("Water Status:","No Water");
      qs.add(water);
      TextQuestion notes = new TextQuestion("Other notes (will be sent to assassins):","");
      qs.add(notes);
      YesNoQuestion correct = new YesNoQuestion("Is this correct?", true);
      qs.add(correct);      
      MenuSelection.askQuestions(qs);
      if (correct.getAnswer())
      {
         Player p = new Player();
         p.setName(name.getAnswer());
         p.setEmail(email.getAnswer());
         p.setCollege(college.getAnswer());
         p.setAddress(address.getAnswer());
         p.setWaterStatus(water.getAnswer());
         p.setNotes(notes.getAnswer());
         System.out.println("Player added succesfully.");
         return p;
      }
      System.out.println("Abort.");
      return null;
   }

   static void editAssassin(Map plugins, Map params) throws AssassinSQLException, IOException
   {
      // StateBuilder sb = new StateBuilder(plugins);
      // try {
      //    state = sb.buildStateAtTime(Tools.getNowTime(), false);
      // } catch (StateBuilderException sbe) {
      //    throw new RuntimeException();
      // }
      List as = Assassin.getAllAssassinsOrderByRealName();
      System.out.println("Assassin to edit?");
      Assassin a = (Assassin) MenuSelection.selectEntity(as);
      if (Debug.debug) Debug.print(autoumpire.class, "got "+a);

      LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
      lsb.registerInterest(a);
      Map state=null;
      try {
         state=lsb.generateState();
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }

      AssassinState ast = (AssassinState) state.get(a);
      if (Debug.debug) Debug.print(autoumpire.class, "got "+ast);

      Event e = new Event();
      e.setActive(true);
      e.setEventTime(Tools.getNowTime());

      // player details
      Player pl = a.getPlayer();
      Vector qs = new Vector();
      TextQuestion name = new TextQuestion("Real Name:", pl.getName());
      qs.add(name);
      TextQuestion email = new TextQuestion("Email:", pl.getEmail());
      qs.add(email);
      TextQuestion college = new TextQuestion("College:", pl.getCollege());
      qs.add(college);
      TextQuestion address = new TextQuestion("Address:", pl.getAddress());
      qs.add(address);
      TextQuestion water = new TextQuestion("Water Status:", pl.getWaterStatus());
      qs.add(water);
      TextQuestion notes = new TextQuestion("Other Notes (will be sent to assassins):", pl.getNotes());
      qs.add(notes);
      YesNoQuestion alive = new YesNoQuestion("Alive?", ast.isAlive());
      qs.add(alive);
      YesNoQuestion active = new YesNoQuestion("Active?", ast.isActive());
      qs.add(active);
      MenuSelection.askQuestions(qs);    
      pl.setName(name.getAnswer());
      pl.setEmail(email.getAnswer());
      pl.setCollege(college.getAnswer());
      pl.setAddress(address.getAnswer());
      pl.setWaterStatus(water.getAnswer());
      pl.setNotes(notes.getAnswer());
      // resurrect them
      if (alive.getAnswer() && !ast.isAlive())
         e.addStateChange(new Resurrection(a));
      // kill them
      if (!alive.getAnswer() && ast.isAlive())
         e.addStateChange(new Death(a, a));
      // deactivate them
      if (!active.getAnswer() && ast.isActive())
         e.addStateChange(new AssassinDeactivation(a));
      // activate them
      if (active.getAnswer() && !ast.isActive())
         e.addStateChange(new AssassinActivation(a));
      
      // plugin details
      List plugs = ast.getKnownPlugins();
      Iterator i = plugs.iterator();
      while (i.hasNext()) {
         String p = (String) i.next();
         if (Debug.debug) Debug.print(autoumpire.class, "got "+p);
         if (!"vanilla".equals(p)) {
            System.out.println("Modifying settings for plugin "+p);
            Map m = ast.getState(p);
            Iterator j = m.keySet().iterator();
            while (j.hasNext()) {
               String key = (String) j.next();
               if (Debug.debug) Debug.print(autoumpire.class, "got "+key);
               Object cval = m.get(key);

               // string
               if (cval instanceof String) {
                  qs = new Vector();
                  TextQuestion question = new TextQuestion("Set "+key,(String) cval);
                  qs.add(question);
                  MenuSelection.askQuestions(qs);    
                  if (!question.getAnswer().equals(cval)) {
                     StateChange sc = new PluginStateChange(a,(Plugin) plugins.get(p), key, question.getAnswer());
                     e.addStateChange(sc);
                  }
               }

               // CanonicalEntity
               else {
                  // NYI!!
               }
            }
         }
      }
   }


   static void addAssassin(Player p2, Map plugins, Map params) throws AssassinSQLException, IOException
   {
      //StateBuilder sb = new StateBuilder(plugins);
      //try {
       //  state = sb.buildStateAtTime(Tools.getNowTime(), false);
      //} catch (StateBuilderException sbe) {
      //   throw new RuntimeException();
        // }

      Player p;
      if (null==p2) {
        List ps = Player.getAllPlayersOrderByName();
      	System.out.println("Player to add assassin for?");
      	p = (Player) MenuSelection.selectEntity(ps);
      } else {
         p=p2;
      }

      Event e = new Event();
      e.setActive(true);
      e.setEventTime(Tools.getNowTime());
      
      Vector qs = new Vector();
      YesNoQuestion active = new YesNoQuestion("Make assassin active?",true);
      qs.add(active);
      TextQuestion pseudonym = new TextQuestion("Pseudonym:", "");
      qs.add(pseudonym);
      MenuSelection.askQuestions(qs);
      
      Assassin a = new Assassin();
      a.setPlayer(p);
      
      if (active.getAnswer())
         e.addStateChange(new AssassinActivation(a));

      Pseudonym ps = null;
      if (!"".equals(pseudonym.getAnswer())) {
         ps = new Pseudonym();
         ps.setValue(pseudonym.getAnswer());
         e.addStateChange(new PseudonymGain(a, ps));
      }
      
      Iterator plgs = plugins.values().iterator();

      LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
      lsb.registerInterest(a);
      Map state=null;
      try {
         state=lsb.generateState();
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      Map statev = viewerifyStateMap(state);
      
      if (Debug.debug) Debug.print(autoumpire.class, "Asking Plugins");
      while (plgs.hasNext()) {
         Plugin pl = (Plugin) plgs.next();
         EventViewer ev = (EventViewer) e.getViewer();
         AssassinViewer av = (AssassinViewer) a.getViewer();
         qs = pl.getQuestions(av, statev);
         if (Debug.debug) Debug.print(autoumpire.class, "Asking "+qs+" for "+pl);
         MenuSelection.askQuestions(qs);
         pl.returnQuestions(av, ev, statev, qs);
      }

      YesNoQuestion correct = new YesNoQuestion("Are these details correct?", true);
      qs.clear();
      qs.add(correct);
      MenuSelection.askQuestions(qs);
      
      if (correct.getAnswer())
      {
         System.out.println("Assassin added succesfully.");
      } else {
         Event.deleteEvent(e);
         Assassin.deleteAssassin(a);
         System.out.println("Abort.");
      }
   }

   static void editEvent(Map plugins, Map params) throws AssassinSQLException, IOException, ParseException
   {
      System.out.println("Select Event to edit: ");
      Event event = (Event) MenuSelection.selectEntity(Event.getAllEventsWithHeadlineOrderByRecentlyModified());
      if (null==event) {
         System.out.println("No Event Selected! Exiting..."); System.exit(1);}
      Vector qs = new Vector();
      TextQuestion time = new TextQuestion("Event Time:", Tools.formatTime(event.getEventTime()));
      qs.add(time);
      String defheadline = event.getHeadline();
      TextQuestion headline = new TextQuestion("Headline:", null==defheadline?"":defheadline);
      qs.add(headline);
      YesNoQuestion active = new YesNoQuestion("Make Active?", event.isActive());
      qs.add(active);
      YesNoQuestion visible = new YesNoQuestion("Make Visible?", event.isVisible());
      qs.add(visible);
      MenuSelection.askQuestions(qs);

      qs = new Vector();
      YesNoQuestion correct = new YesNoQuestion("Is this correct?", true);
      qs.add(correct);
      MenuSelection.askQuestions(qs);

      if (correct.getAnswer()) {
         System.out.println("Event added succesfully.");
         event.setEventTime(Tools.unFormatTime(time.getAnswer()));
         event.setHeadline(headline.getAnswer());
         event.setActive(active.getAnswer());
         event.setVisible(visible.getAnswer());
         Event.commitAll(); // Do we need this? I'm not sure, but I figure we must
      } else {
         System.out.println("Abort.");
      }
   }
    
   static Event addEvent(Map plugins, Map params) throws AssassinSQLException, IOException, ParseException
   {
         //      Mike's Crazy Mods
         //      StateBuilder sb = new StateBuilder(plugins);
         Map myState=null; // new
         //      try {
         //         state = sb.buildStateAtTime(Tools.getNowTime(), false);
         //      } catch (StateBuilderException sbe) {
         //         throw new RuntimeException();
         //      }
         //      if (Debug.debug) Debug.printMap(autoumpire.class, state);
         
      List as = Assassin.getAllAssassinsOrderByRealName();
      System.out.println("Select the assassins involved in the event: ");
      Event e = new Event();
      long now = Tools.getNowTime(); // new
      LightweightStateBuilder lsb = new LightweightStateBuilder(now); // new
      Assassin a = (Assassin) MenuSelection.selectEntity(as);
      while (null != a) {
      //         if (Debug.debug) Debug.print(autoumpire.class, "Selected State: " + state.get(a));
         lsb.registerInterest(a);
         a = (Assassin) MenuSelection.selectEntity(as);
      }
      try {
         myState = lsb.generateState(); // new -- This mings, but it's the way we have to do it
      } catch (StateBuilderException sbe) {
         // WIBBLE!
         if (Debug.debug) Debug.print(autoumpire.class, "State Builder Exception: "+sbe.getMessage());
         throw new RuntimeException();
      }
      int n = 1;
      for (Iterator i = lsb.getInvolved().iterator(); i.hasNext();) {
         Assassin a2 = (Assassin) i.next();
         System.out.println("Select pseudonym for "+a2.getPlayer().getName());
         Pseudonym p = (Pseudonym) MenuSelection.selectEntity(((AssassinState) myState.get(a2)).getPseudonyms()); // new
         e.setParticipant(n++, a2, p);
      }
        //          p = (Pseudonym) MenuSelection.selectEntity(((AssassinState) state.get(a)).getPseudonyms());
        
      Vector qs = new Vector();
      TextQuestion time = new TextQuestion("Event Time:", Tools.formatTime(Tools.getNowTime()));
      qs.add(time);
      String defheadline = Config.getConfig("vanilla", "default-headline");

      Map m = e.getParticipants();
      for (Iterator i = m.keySet().iterator(); i.hasNext(); ) {
	      Integer j = (Integer) i.next();
	      Assassin part = (Assassin) m.get(j);
	      System.out.println("("+j+"): "+part.getPrintable());
      }

      TextQuestion headline = new TextQuestion("Headline:", null==defheadline?"":defheadline);
      qs.add(headline);
      
      YesNoQuestion active = new YesNoQuestion("Make Active?", true);
      qs.add(active);
      
      YesNoQuestion visible = new YesNoQuestion("Make Visible?", true);
      qs.add(visible);

      MenuSelection.askQuestions(qs);

      e.setEventTime(Tools.unFormatTime(time.getAnswer()));
      e.setHeadline(headline.getAnswer());
      e.setActive(active.getAnswer());
      e.setVisible(visible.getAnswer());
     
      System.out.println("Enter Deaths:");
      do {
         System.out.println("Select victim: ");
         Assassin victim = (Assassin) MenuSelection.selectEntity(new Vector(e.getParticipants().values()));
         if (null == victim) break;
         System.out.println("Select killer: ");
         Assassin killer = (Assassin) MenuSelection.selectEntity(new Vector(e.getParticipants().values()));
         if (null == killer) break;
         e.addStateChange(new Death(killer, victim));
      } while (true);
      
      Iterator ps = plugins.values().iterator();
      Map statev = viewerifyStateMap(myState);
      
      while (ps.hasNext()) {
         Plugin p = (Plugin) ps.next();
         EventViewer ev = (EventViewer) e.getViewer();
         qs = p.getQuestions(ev, statev);
         MenuSelection.askQuestions(qs);
         p.returnQuestions(ev, statev, qs);
      }

      qs = new Vector();
      YesNoQuestion correct = new YesNoQuestion("Is this correct?", true);
      qs.add(correct);
      MenuSelection.askQuestions(qs);

      if (correct.getAnswer())
      {
         System.out.println("Event added succesfully.");
         return e;
      }
      // Else
      Event.deleteEvent(e);
      System.out.println("Abort.");
      return null;
   }
    
   static void editReport(Map plugins, Map params) throws AssassinSQLException, IOException
   {
      System.out.println("Select Report: ");
      Report rep = (Report) MenuSelection.selectEntity(Report.getAllReports());
      if (null == rep) {
         System.out.println("No report selected, exiting"); System.exit(1); }
      System.out.println("Select Reporting Assassin: ");
      Assassin assassin = (Assassin) MenuSelection.selectEntity(new Vector(rep.getEvent().getParticipants().values()));
      Pseudonym pseudonym = null;
      if (null == assassin) 
         System.out.println("Adding Comment");
      else {
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         lsb.registerInterest(assassin);
         Map state=null;
         try {
            state=lsb.generateState();
         } catch (StateBuilderException sbe) {
            throw new RuntimeException();
         }
         System.out.println("Select Pseudonym for Assassin: ");
         pseudonym = (Pseudonym) MenuSelection.selectEntity(((AssassinState) state.get(assassin)).getPseudonyms());
      }

      Random r = new Random();
      String file = "/tmp/tmp.ass."+r.nextInt();
      (new File(file)).deleteOnExit();
      
      execnative("touch \""+file+"\"; chmod 600 \""+file+"\"");
      
      PrintWriter out = new PrintWriter(new FileOutputStream(file));
      out.println(rep.getText());
      out.close();

      execnative("if [[ $EDITOR == \"\" ]] ; then nano \""+file+"\" ; else $EDITOR \""+file+"\" ; fi");

      String text = "";
      BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
      String s;
      while (null != (s = in.readLine()))
         text += s+"\n";
      in.close();

      YesNoQuestion correct = new YesNoQuestion("Correct?", true);
      Vector qs = new Vector();
      qs.add(correct);
      MenuSelection.askQuestions(qs);
      if (correct.getAnswer())
      {
         if (Debug.debug) Debug.print(autoumpire.class, "Set Assassin "+assassin);
         rep.setAssassin(assassin);
         if (Debug.debug) Debug.print(autoumpire.class, "Set Pseudonym "+pseudonym);
         rep.setPseudonym(pseudonym);
         if (Debug.debug) Debug.print(autoumpire.class, "Set Text "+text);
         rep.setText(text);
         if (Debug.debug) Debug.print(autoumpire.class, "Write Out");
         System.out.println("Report updated succesfully.");
         Report.commitAll();
      } else
         System.out.println("Abort.");
   }

   static void addReport(Event e, Map plugins, Map params) throws AssassinSQLException, IOException
   {
      //StateBuilder sb = new StateBuilder(plugins);
      //try {
      //   state = sb.buildStateAtTime(new Date().getTime(), false);
      //} catch (StateBuilderException sbe) {
      //   throw new RuntimeException();
      //}

      String repFile = (String) params.get("-r");
      String report = "";

      if (null != repFile) {
         File f = new File(repFile);
         InputStream in = new BufferedInputStream(new FileInputStream(f));
         while (in.available() > 0)
            report += (char) in.read();
         in.close();
      }

      BufferedReader inp = new BufferedReader(new InputStreamReader(System.in));
      
      Event event;
      if (null==e) {
         System.out.println("Select Event to add report to: ");
         event = (Event) MenuSelection.selectEntity(Event.getAllEventsWithHeadlineOrderByRecentlyModified());
      }  else {
         event = e;
      }
      if (null == event) {
         System.out.println("No event selected, exiting"); System.exit(1); }
      System.out.println("Select Reporting Assassin: ");
      Assassin assassin = (Assassin) MenuSelection.selectEntity(new Vector(event.getParticipants().values()));
      Pseudonym pseudonym = null;
      if (null == assassin) 
         System.out.println("Adding Comment");
      else {
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         lsb.registerInterest(assassin);
         Map state=null;
         try {
            state=lsb.generateState();
         } catch (StateBuilderException sbe) {
            throw new RuntimeException();
         }
         System.out.println("Select Pseudonym for Assassin: ");
         pseudonym = (Pseudonym) MenuSelection.selectEntity(((AssassinState) state.get(assassin)).getPseudonyms());
      }
      if ("".equals(report)) {

         if (natv) {
            Random r = new Random();
            String file = "/tmp/tmp.ass."+r.nextInt();
            (new File(file)).deleteOnExit();

            execnative("touch \""+file+"\"; chmod 600 \""+file+"\"");

            execnative("if [[ $EDITOR == \"\" ]] ; then nano \""+file+"\" ; else $EDITOR \""+file+"\" ; fi");

            BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
            report = "";
            String s;
            while (null != (s = in.readLine()))
               report += s+"\n";
            in.close();
         } 
         else {
            System.out.println("Enter Report Text (end with ^D):");
            String temp = "";
            do {
               report += temp + "\n";
               temp = inp.readLine();
            } while (null != temp);
         }
      } else {
         System.out.println("Report Text:");
         System.out.println(report);
      }
      YesNoQuestion correct = new YesNoQuestion("Add report?", true);
      Vector qs = new Vector();
      qs.add(correct);
      MenuSelection.askQuestions(qs);
      if (correct.getAnswer())
      {
         Report r = new Report();
         if (Debug.debug) Debug.print(autoumpire.class, "Set Assassin "+assassin);
         r.setAssassin(assassin);
         if (Debug.debug) Debug.print(autoumpire.class, "Set Pseudonym "+pseudonym);
         r.setPseudonym(pseudonym);
         if (Debug.debug) Debug.print(autoumpire.class, "Set Text "+report);
         r.setText(report);
         if (Debug.debug) Debug.print(autoumpire.class, "Add To Event "+event);
         event.addReport(r);
         if (Debug.debug) Debug.print(autoumpire.class, "Write Out");
         System.out.println("Report added succesfully.");
         Event.commitAll();
      } else
         System.out.println("Abort.");
   }

   static void addPseudonym(Map plugins, Map params) throws AssassinSQLException, IOException
   {
      System.out.println("Select Assassin: ");
      Assassin assassin = (Assassin) MenuSelection.selectEntity(Assassin.getAllAssassinsOrderByRealName());
      
      Vector qs = new Vector();
      TextQuestion pseudonym = new TextQuestion("Enter Pseudonym:", "");
      qs.add(pseudonym);
      YesNoQuestion correct = new YesNoQuestion("Is this correct?", true);
      qs.add(correct);
      MenuSelection.askQuestions(qs);
      
      if (correct.getAnswer())
      {
         Pseudonym p = new Pseudonym();
         p.setValue(pseudonym.getAnswer());
         Event e = new Event();
         e.setActive(true);
         e.setEventTime(Tools.getNowTime());
         e.addStateChange(new PseudonymGain(assassin, p));
         System.out.println("Pseudonym added succesfully.");
      } else
         System.out.println("Abort.");
   }

   static void updateState(Map plugins, Map params) throws AssassinSQLException, IOException, ParseException
   {
      StateBuilder sb = new StateBuilder(plugins);
      Map state;
      long oldtime;
      long time = Tools.getNowTime();
      try {
         state = sb.buildStateAtTime(time);
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      if (Debug.debug) Debug.print(autoumpire.class, "updating state");
      Iterator pni = plugins.keySet().iterator();
      while (pni.hasNext())
      {
         String pname = (String) pni.next();
         if (Debug.debug) Debug.print(autoumpire.class, "Plugin name is " + pname);
         Plugin p = (Plugin) plugins.get(pname);
         if (Debug.debug) Debug.print(autoumpire.class, "Updating state for " + pname);
         try
         {
            Map statev = viewerifyStateMap(state);
            Event e = new Event();
            e.setEventTime(Tools.getNowTime());
            e.setModifiedTime(Tools.getNowTime());
            e.setActive(true);
            e.setVisible(false);
            if (Debug.debug) Debug.print(autoumpire.class, "calling updateState");
            Vector v1 = p.updateState(statev, new Vector(), (EventViewer) e.getViewer());
            Vector v2 = v1;
            if (Debug.debug) Debug.print(autoumpire.class, "updateState returned " + v2);
            while (v1 != null)
            {
               if (Debug.debug) Debug.print(autoumpire.class, "question vector is " + v2.toString());
               sb = new StateBuilder(plugins);
               oldtime = time;
               time = Tools.getNowTime();
               state = sb.reBuildStateAtTime(oldtime, state, time);
               statev = viewerifyStateMap(state);
               MenuSelection.askQuestions(v1);
               v2.addAll(v1);
               v1 = p.updateState(statev, v2, (EventViewer) e.getViewer());
               Event.commitAll();
               StateChange.commitAll();
            }
            if (e.getStateChanges().size() == 0) {
               Event.deleteEvent(e);
               e = null;
            }
         }
         catch (StateBuilderException SBe)
         {
            System.err.println("autoumpire: Error while building state.");
            System.err.println(SBe.getMessage());
            System.exit(1);
         }
      }
   }


   static void generateMail(Map plugins, Map params)  throws AssassinSQLException, IOException, ParseException
   {
      StateBuilder sb = new StateBuilder(plugins);
      Map state;
      try {
         state = sb.buildStateAtTime(Tools.getNowTime());
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      boolean force = false;
      String textFile = "";
      String outFile = "";
      force = params.containsKey("-a");
      textFile = (String) params.get("-e");
      outFile = (String) params.get("-o");

      if (null == outFile) outFile = "email";

      if (Debug.debug) Debug.print(autoumpire.class, "Arguments to generate-mail: force: "+force);
      String manualText = "";
      if (null != textFile) {
         File f = new File(textFile);
         InputStream in = new BufferedInputStream(new FileInputStream(f));
         while (in.available() > 0)
            manualText += (char) in.read();
         in.close();
      }
      String emails = sendEmails(state, plugins, force, manualText);
      File o = new File(outFile);
      PrintWriter out = new PrintWriter(new FileOutputStream(o));
      out.println(emails);
      out.flush();
      out.close();
   }

   static void generatePages(Map plugins, Map params)  throws AssassinSQLException, IOException, ParseException, StateBuilderException, XMLParseException
   {
      if (Debug.debug) Debug.print(autoumpire.class, "Generating pages");
      StateBuilder sb = new StateBuilder(plugins);
      Map state = sb.doNews(Tools.getNowTime());
      // need to run xsltproc on all the news....
      String path = Config.getConfig("vanilla","path-to-pages");
      String xslt_template = Config.getConfig("vanilla","xsl-file");
      
      /* Iterating over all the desired templates */
      int i = 1;
      String myTemplate = Config.getConfig("vanilla","template-" + i);
      while (null != myTemplate) {
         String out = Config.getConfig("vanilla","template-page-name-" + i);
         if (Debug.debug) Debug.print(autoumpire.class, "Path "+path+", Template "+myTemplate+", out "+out);

         // generate the XML
         WebPageParser wl = new WebPageParser(myTemplate,state,plugins,path+out+".xml",sb.getPages());
         wl.generate();
         
         // convert to HTML
         if (Debug.debug) Debug.print(autoumpire.class,"exec(/usr/bin/xsltproc -o "+path+out+".html "+xslt_template+" "+path+out+".xml)");
         Tools.exec(new String[] {
            "/usr/bin/xsltproc",
            "--novalid",
            "-o",
            path+out+".html",
            xslt_template,
            path+out+".xml"});
         
         myTemplate = Config.getConfig("vanilla","template-" + ++i);
      }
      if (Debug.debug) Debug.print(autoumpire.class, "Generating done");
   }
   static void dumpKillTree(Map plugins, Map params) throws AssassinSQLException {
      List deathlist = StateChange.getByClassAtTime(Death.class, Tools.getNowTime());
      System.out.println("digraph KillTree {");
      for (Iterator i = deathlist.iterator(); i.hasNext();) {
         Death d = (Death) i.next();
         AssassinViewer k = d.getKiller();
         AssassinViewer v = d.getVictim();
         System.out.println("\"" + k.getPlayerViewer().getName() + "\" -> \"" + v.getPlayerViewer().getName() + "\";");
      }
      System.out.println("}");
   }
   
   static void addParticipant(Map plugins, Map params) throws IOException, AssassinSQLException {
      System.out.println("Select Event to add participant to: ");
      Event event = (Event) MenuSelection.selectEntity(Event.getAllEventsWithHeadlineOrderByRecentlyModified());
      if (null == event) {
         System.out.println("No event selected, exiting"); System.exit(1); }
      System.out.println("Select Assassin to add: ");
      Assassin assassin = (Assassin) MenuSelection.selectEntity(Assassin.getAllAssassinsOrderByRealName());
      if (null == assassin) {
         System.out.println("No assassin selected, exiting"); System.exit(1); }
      LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
      lsb.registerInterest(assassin);
      Map state=null;
      try {
         state=lsb.generateState();
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      System.out.println("Select Assassin to add: ");
      Pseudonym pseudo = (Pseudonym) MenuSelection.selectEntity(((AssassinState)state.get(assassin)).getPseudonyms());
      if (null==pseudo) {
         System.out.println("No pseudonym selected, exiting"); System.exit(1); }
      int m = event.getMaxParticipantID();
      event.setParticipant(++m, assassin, pseudo);
   }
}
