/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.pistachio;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.ConfigProperty;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginCommand;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.TextQuestion;
import org.ucam.srcf.assassins.YesNoQuestion;
import org.ucam.srcf.assassins.autoumpire;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Pistachio extends Plugin
{
   public static String getDescription() { return "Police plugin, designate as police, give ranks, etc."; };
   private String defaultRank;
   private String defNumRank;
   public void setPluginName(String s) 
   {
      super.setPluginName(s);
      defaultRank = Config.getConfig(getPluginName(), "default-rank");
      defNumRank = Config.getConfig(getPluginName(), "default-num-rank");
   }
   
   /**
    * Called by the filtering mechanism when the plugin needs to determine
    * whether an assassin has a given property such as &quot;police&quot;. If
    * the plugin does not recognise the named property, it should return
    * <code>false</code>.
    * @param a A viewer for the assassin being tested.
    * @param prop The name of the property being asked about.
    * @return True if the assassin has the property, false if not.
    */
   public boolean assassinHasProperty(AssassinStateViewer asv, String prop, String value)
   {
      if (Debug.debug) Debug.print(autoumpire.class, "assassinHasProperty("+asv+","+prop+","+value+")");
      if (!"police".equals(prop)) return false;
      String police = asv.getProperty(this, "police");
      if (null == police) return false;
      if (police.equals(value)) return true;
      return false;
   }

   /**
    * Called when the system has decided that email needs to be sent to a given
    * player. The plugin may return a string to be included in the email, or
    * null if it does not wish to include anything.
    * @param p A viewer for the player being emailed.
    * @return A string to be included in the email sent to the player, or null.
    *    The string will appear in its own paragraph when inserted into the
    *    email and so should not begin or end with a newline.
    */
   protected String generateEmail(AssassinStateViewer asv) 
   {
      return "";
   }

   /**
    * Called when an assassin is being added to get any queries the plugin
    * has for the user. The plugin should check all the assassin details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param a The Assassin which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector getQuestions(AssassinViewer a, Map state) throws AssassinSQLException 
   {
      Vector qs = new Vector();
      qs.add(new YesNoQuestion("Make Police",false));
      qs.add(new TextQuestion("Rank (textual)",defaultRank));
      qs.add(new TextQuestion("Rank (numeric)",defNumRank));
      return qs;
   }
   
   /**
    * Returns the Questions with the answers filled out
    *
    * @param a The Assassin being added
    * @param qs The Questions which were asked
    */ 
   protected void returnQuestions(AssassinViewer a, EventViewer e, Map state, Vector qs) throws AssassinSQLException
   {
      boolean police = ((YesNoQuestion)qs.get(0)).getAnswer();      
      vanilla.addStateChange(e, new PluginStateChange(a, this, "police", police?"true":"false"));
      String rank = ((TextQuestion)qs.get(1)).getAnswer();
      vanilla.addStateChange(e, new PluginStateChange(a, this, "rank", rank));
      String numrank = ((TextQuestion)qs.get(2)).getAnswer();
      vanilla.addStateChange(e, new PluginStateChange(a, this, "numrank", numrank));
   }

   private void printPolice(Map params) throws AssassinSQLException, StateBuilderException
   {
      Map state = vanilla.getState();
      Iterator i = state.values().iterator();
      while (i.hasNext()) {
         AssassinStateViewer asv = (AssassinStateViewer) i.next();
         if (asv.isAlive() && asv.isActive() && asv.getBoolProperty(this, "police")) {            
            System.out.print(asv.getProperty(this, "rank"));
            System.out.print(" ");
            System.out.print(asv.getAssassin().getPlayerViewer().getName());
            System.out.println("");
         }
      }
   }

   /**
    * Returns a Map of new commands to add to the CLI
    *
    * @return A Map of String =&gt; Command
    */
   public Map getCommands()
   { 
      Map m = new HashMap();
      final Pistachio pist = this;
      m.put("print-police",
             new PluginCommand () {
               public void pluginCommand(Map params) throws AssassinSQLException, StateBuilderException
               {
                  pist.printPolice(params);
               }  
            }
            );
      return m;
   }
   
   public  static Vector getConfigProperties() 
   { 
      Vector v = new Vector(); 
      v.add ( new ConfigProperty ("default-rank", "Default Rank (textual)", "Constable"));
      v.add ( new ConfigProperty ("default-num-rank", "Default Rank (numerical)", "0"));
      return v;
   }

}


